extern crate chrono;
extern crate which;

use chrono::prelude::*;
use std::process::Command;

fn main() {
    // Revision Info
    set_version_env_vars();
    // Build Time + Date
    set_build_timedate();
}

/// Sets version environment vars from the Version Control System
///
/// **Env Vars:**
///
///```no_run
/// VCS_COMMIT_DATE
/// VCS_COMMIT_HASH
///```
fn set_version_env_vars() {
    /// Supported Version Control Systems
    #[derive(Debug, PartialEq, Eq)]
    enum VCS {
        Git,
        Mercurial,
        Pijul,
        Fossil,
        /// Fallback for no version control
        None,
    }

    impl VCS {
        /// Returns the executable name for the VCS system
        fn binary_name(&self) -> &str {
            match self {
                Self::Git => "git",
                Self::Mercurial => "hg",
                Self::Pijul => "pijul",
                Self::Fossil => "fossil",
                Self::None => "",
            }
        }
    }

    fn vcs_from_str(s: &str) -> VCS {
        match s {
            ".git" => VCS::Git,
            ".hg" => VCS::Mercurial,
            ".pijul" => VCS::Pijul,
            ".fossil" => VCS::Fossil,
            _ => VCS::None,
        }
    }

    use std::fs::read_dir;
    use which::which;
    /////////////////////////////////////////////////////////////////////////
    // Do the needful here.
    /////////////////////////////////////////////////////////////////////////
    // Figure out what Version Control System we're using.
    // Get all directories in the project root
    let dirs = read_dir(env!("CARGO_MANIFEST_DIR")).unwrap();
    let mut vcs = VCS::None; // Default if we don't find any VCS folders.
                             // Search until we find a VCS folder.
    for folder in dirs {
        let v = vcs_from_str(folder.unwrap().file_name().into_string().unwrap().as_str());
        if v != VCS::None {
            vcs = v;
            break; // Go with the first VCS folder we find.
        }
    }

    // Variables to pass into env
    let mut commit_date = "0000-00-00".to_string(); // Defaults
    let mut commit_hash = "ffffffffffff".to_string(); // Defaults

    let bin: &str = vcs.binary_name();

    // Replace the default values below with the correct versions from VCS
    match vcs {
        VCS::None => {
            // We can't pull any information, use the Defaults assigned above.
            println!("cargo:warning=Unable to detect Version Control System, program wil not have correct build info!");
        }
        VCS::Git => {
            match which(bin) {
                Ok(_val) => {
                    let stdout = String::from_utf8(
                        Command::new(bin)
                            .args(&["log", "-1", "--pretty=%as%n%H"])
                            .output()
                            .unwrap()
                            .stdout,
                    )
                    .unwrap();
                    let output = stdout.lines().collect::<Vec<&str>>();
                    commit_date = output[0].trim().to_owned();
                    // Git commit hashes have a short UID of 7 characters
                    // The above placeholder is formatted for 12 characters for
                    // Mercurial shorthash standards.
                    // Both systems have a 40 character "full" hash. 40 can be
                    // trimmed down to 12, and git can recognize the commit based
                    // on just those first 7+ anyway (and in some cases, less.)
                    commit_hash = output[1][..12].trim().to_owned();
                }
                Err(_err) => {
                    println!("cargo:warning=Attempt to use git for version information, but binary does not exist.");
                }
            }
        }
        VCS::Mercurial => match which(bin) {
            Ok(_val) => {
                let stdout = String::from_utf8(
                    Command::new(bin)
                        .args(&[
                            "log",
                            "--rev",
                            ".",
                            "--template",
                            "{date|shortdate}\n{node|short}",
                        ])
                        .output()
                        .unwrap()
                        .stdout,
                )
                .unwrap();
                let output = stdout.lines().collect::<Vec<&str>>();
                commit_date = output[0].trim().to_owned();
                commit_hash = output[1].trim().to_owned();
            }
            Err(_err) => {
                println!("cargo:warning=Attempt to use Mercurial for version information, but binary does not exist.");
            }
        },
        VCS::Pijul => {
            todo!("Pijul version information not implemented!!")
        }
        VCS::Fossil => {
            todo!("Fossil version information not implemented!!")
        }
    }
    // Output the build information from version control to Cargo envs
    println!("cargo:rustc-env=VCS_COMMIT_DATE={}", commit_date);
    println!("cargo:rustc-env=VCS_COMMIT_HASH={}", commit_hash);
}

/// Sets environment vars for when the binary was built
///
/// **Env Vars:**
///
///```no_run
/// BUILD_TIMEDATE
///```
fn set_build_timedate() {
    let utc: DateTime<Utc> = Utc::now();
    // Output the build information to Cargo envs
    println!("cargo:rustc-env=BUILD_TIMEDATE={}", utc.to_rfc2822());
}
