#![allow(non_snake_case)] // Because taken from JS

pub static API_URL: &str = "https://api.knockout.chat/v2/";

use std::{fs::File, io::BufWriter, sync::Arc, time::Duration};

use chrono::{DateTime, Utc};
use core::sync::atomic::AtomicU64;
use futures::StreamExt;
use reqwest::Client;
use serde::{Deserialize, Deserializer, Serialize};
use std::io::Write;

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct Ban {
    pub id: u64,
    pub banReason: String,
    pub expiresAt: DateTime<Utc>,
    pub createdAt: DateTime<Utc>,
    pub updatedAt: DateTime<Utc>,
    pub post: u64,
    pub user: User,
    pub bannedBy: User,
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct User {
    #[serde(deserialize_with = "deserialize_null_default")]
    pub id: u64,
    pub role: Role,
    #[serde(deserialize_with = "deserialize_null_username")]
    pub username: String,
    pub avatarUrl: Option<String>,
    pub backgroundUrl: Option<String>,
    pub posts: u64,
    pub threads: u64,
    pub createdAt: DateTime<Utc>,
    pub updatedAt: DateTime<Utc>,
    pub banned: bool,
    pub isBanned: bool,
    #[serde(deserialize_with = "deserialize_nasty_utf8")]
    pub title: Option<String>,
    #[serde(deserialize_with = "deserialize_null_default")]
    pub pronouns: Option<String>,
    #[serde(deserialize_with = "deserialize_null_default")]
    pub useBioForTitle: Option<u8>,
}

fn deserialize_null_username<'de, D>(deserializer: D) -> Result<String, D::Error>
where
    D: Deserializer<'de>,
{
    let opt = Option::deserialize(deserializer).unwrap_or(Some(String::from("Deleted User")));
    match opt {
        Some(v) => Ok(v),
        None => Ok(String::from("Deleted User")),
    }
}

fn deserialize_nasty_utf8<'de, D>(deserializer: D) -> Result<Option<String>, D::Error>
where
    D: Deserializer<'de>,
{
    let opt: Result<Option<String>, D::Error> = Option::deserialize(deserializer);
    Ok(opt.unwrap_or_else(|_| Some(String::new())))
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct Role {
    pub id: u64,
    pub code: String,
    pub description: String,
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct Rating {
    pub id: u64,
    pub rating: String,
    pub ratingId: u64,
    pub users: Vec<User>,
    pub count: u64,
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct Post {
    pub id: u64,
    #[serde(alias = "threadId")]
    pub thread: u64,
    pub page: u64,
    #[serde(deserialize_with = "deserialize_null_default")]
    pub content: String,
    pub createdAt: DateTime<Utc>,
    pub updatedAt: DateTime<Utc>,
    pub user: User,
    pub ratings: Vec<Rating>,
    pub bans: Vec<Ban>,
    pub threadPostNumber: u64,
    pub countryName: Option<String>,
    pub countryCode: Option<String>,
    pub appName: Option<String>,
}



fn deserialize_null_default<'de, D, T>(deserializer: D) -> Result<T, D::Error>
where
    T: Default + Deserialize<'de>,
    D: Deserializer<'de>,
{
    let opt = Option::deserialize(deserializer)?;
    Ok(opt.unwrap_or_default())
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct Thread {
    pub id: u64,
    pub title: String,
    #[serde(deserialize_with = "deserialize_null_default")]
    pub iconId: u64,
    pub subforumId: u64,
    pub createdAt: DateTime<Utc>,
    pub updatedAt: DateTime<Utc>,
    pub deletedAt: Option<String>,
    pub deleted: bool,
    pub locked: bool,
    pub pinned: bool,
    pub lastPost: Post,
    pub backgroundUrl: Option<String>,
    pub backgroundType: Option<String>,
    pub user: User,
    pub postCount: u64,
    pub subforum: Subforum,
    pub totalPosts: u64,
    pub currentPage: u64,
    pub threadBackgroundUrl: Option<String>,
    pub threadBackgroundType: Option<String>,
    pub posts: Vec<Post>,
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct Subforum {
    pub id: u64,
    pub name: String,
    pub description: String,
    #[serde(deserialize_with = "deserialize_null_default")]
    pub iconId: u64,
    pub icon: String,
}
/// Writes the JSON and associated error to the given path
fn write_error(path: String, body: &[u8], err: String) -> Result<(), std::io::Error> {
    let file = File::create(path)?;
    let mut fbuf = BufWriter::new(file);
    fbuf.write_all(body)?;
    write!(fbuf, "\n{}", err)?;
    Ok(())
}

#[derive(Debug)]
enum ThreadGetError {
    Retry(eyre::Error),
    Abort(eyre::Error),
}

impl From<eyre::Error> for ThreadGetError {
    fn from(err: eyre::Error) -> Self {
        ThreadGetError::Abort(err)
    }
}

fn test_resp(resp: reqwest::Response) -> Result<reqwest::Response, ThreadGetError> {
    let resp_code = resp.status();
    // This means it was an error on our part, probably a bad request
    // Meaning we shouldn't retry, as opposed to other errors.
    (!resp_code.is_client_error())
        .then(|| resp)
        .ok_or_else(|| ThreadGetError::Abort(eyre::Report::msg(resp_code)))
}

async fn get_thread(
    client: &Client,
    thread_url: &String,
    id: u64,
) -> Result<Thread, ThreadGetError> {
    use ThreadGetError::*;
    let resp = client
        .get(thread_url.clone())
        .send()
        .await
        .map_err(|e| Retry(e.into()))
        .and_then(|e| test_resp(e))?;

    // Get the text from our known good response, and deserialize it.
    let resp_body = resp.bytes().await.unwrap();
    if resp_body.as_ref() == b"Application error. Please try again later" {
        return Err(Retry(eyre::Report::msg(
            "Application error. Please try again later",
        )));
    }

    let deser = serde_json::from_slice(&resp_body);
    match deser {
        Ok(data) => {
            return Ok(data);
        }
        Err(err) => {
            let path = format!("threads/{}.err.json", id,);
            write_error(path, &resp_body, err.to_string()).unwrap();
            return Err(Abort(err.into()));
        }
    }
}

async fn get_page(
    client: &Client,
    in_thread: &Thread,
    thread_url: &String,
) -> Result<Vec<Result<Vec<Post>, eyre::Error>>, eyre::Error> {
    // We got the thread information, so now we can beging scraping pages.
    let total_pages = in_thread.lastPost.page; // Total Pages in thread.
    let last_post_num = in_thread.lastPost.threadPostNumber as usize;
    let scrape_atomic = &Arc::new(AtomicU64::new(0));

    println!("Thread:\t{}\t\"{}\"", in_thread.id, in_thread.title);
    println!("Total Pages: {}", total_pages);

    // Builds our page URLs into a vector
    // This is later passed into an async unordered buffer executor.
    let mut urls: Vec<String> = Vec::with_capacity(total_pages as usize);
    for n in 1..=total_pages {
        urls.push(format!("{}/{}", thread_url.clone(), n));
    }

    // Collect all the pages into an unordered buffer.
    let fetches = futures::stream::iter(urls.into_iter().map(|url| {
        async move {
            let mut iters = 0;
            loop {
                iters = iters + 1; // For backpressure.
                                   // Slow down a bit to not overwhelm the server.
                tokio::time::sleep(Duration::from_millis(10)).await;

                let thread_str; // Assigned in Ok client.get below.
                let mut tries = 0;
                loop {
                    if tries >= 3 {
                        return Err(eyre::Report::msg(format!(
                            "Timed out trying connection to {}",
                            url.clone()
                        )));
                    }
                    match client.get(url.clone()).send().await {
                        Ok(res) => match res.bytes().await {
                            Ok(v) => {
                                thread_str = v;
                                break;
                            }
                            Err(e) => {
                                println!(
                                    "\t Failed to read body: {}\n\t  Reason: {}",
                                    url.clone(),
                                    e
                                )
                            }
                        },
                        Err(err) => {
                            println!("\t Failed to connect: {}\n\t  Reason: {}", url.clone(), err);
                        }
                    }
                    tries += 1;
                }

                let thread: Thread; // No initial value, assign inside following match.
                match serde_json::from_slice(&thread_str) {
                    Ok(deser) => {
                        thread = deser;
                    }
                    Err(_err) => {
                        if _err.to_string().contains("invalid type") {
                            println!("Page: {}", url);
                            let path = format!(
                                "threads/{}-{}.err.json",
                                in_thread.id,
                                url.clone().split("/").last().unwrap()
                            );
                            println!("\t DESER ERROR: {}", _err);
                            println!("\t Wrote error to {}", path);
                            write_error(path, &thread_str, _err.to_string()).unwrap();
                            return Err(eyre::Report::msg("Deserialization Errors".to_string()));
                        }
                        if iters >= 25 {
                            return Err(eyre::Report::msg(_err.to_string()));
                        }
                        // Probably just overloaded the server, hold up a sec.
                        tokio::time::sleep(Duration::from_millis(iters * 25)).await;
                        continue; // Restart on this request.
                    }
                }
                // Increment our progress
                let mut progress = scrape_atomic.load(std::sync::atomic::Ordering::Acquire);
                progress = progress + 1;
                scrape_atomic.store(progress, std::sync::atomic::Ordering::Release);
                // Build a list of users.
                //let users = build_users(thread.posts);
                // Display our progress
                println!(
                    "Processed:... {:width$}/{page} | Page: {realpage:width$}",
                    progress,
                    width = format!("{}", last_post_num).chars().count(),
                    page = thread.lastPost.page,
                    realpage = thread.posts.last().unwrap().page,
                );
                break Ok(thread.posts);
            }
        }
    }))
    // KO accepts up to 25 connections, we already utilized 1 for the thread
    .buffer_unordered((total_pages as usize).min(25).max(1))
    .collect::<Vec<Result<Vec<Post>, eyre::Error>>>()
    .await;
    Ok(fetches)
}

pub async fn scrape_thread(client: &Client, id: u64) -> Result<Thread, eyre::Error> {
    let thread_url = format!("{}threads/{}", API_URL, id);
    let mut in_thread: Thread; // Will be allocated inside the loop.

    let mut iters: u64 = 0;
    loop {
        /////////////////////////////////////////////////////////////////
        // Connection Loop for 'in_thread'.
        // We use this to catch all the HTTP and deserilization errors we
        // might get on attempting to get the thread info.
        // Failing to get the thread info means we can't get any pages,
        // and will return an Error() as such.
        /////////////////////////////////////////////////////////////////
        iters = iters + 1; // Increment our iteration count.

        match get_thread(&client, &thread_url, id).await {
            Ok(v) => {
                in_thread = v;
                break;
            }
            Err(e) => match e {
                ThreadGetError::Retry(e) => {
                    // Try, try, try again.
                    if iters >= 10 {
                        return Err(e);
                    }
                    tokio::time::sleep(Duration::from_millis(iters * 25)).await;
                    continue;
                }
                ThreadGetError::Abort(e) => {
                    return Err(e);
                }
            },
        }
    }
    let fetches = get_page(&client, &in_thread, &thread_url).await?;
    // Merge our unordered buffer into a straight Vec.
    let mut post_buffer: Vec<Post> = Vec::with_capacity(in_thread.postCount as usize);
    for vec in fetches.into_iter() {
        match vec {
            Ok(mut posts) => {
                post_buffer.append(&mut posts);
            }
            // On first encounter of an error, send it back to the calling function
            Err(err) => {
                return Err(eyre::Report::msg(err));
            }
        }
    }
    // Sort our Post and UserDocument Vecs.
    post_buffer.sort_by(|a, b| a.threadPostNumber.cmp(&b.threadPostNumber));
    // Move all posts from all pages into the original thread struct.
    in_thread.posts = post_buffer;
    Ok(in_thread)
}
