pub const AUTHORS: &str = env!("CARGO_PKG_AUTHORS");
pub const VERSION: &str = env!("CARGO_PKG_VERSION");
pub const BUILDTD: &str = env!("BUILD_TIMEDATE");
pub const REVISION: &str = env!("VCS_COMMIT_HASH");
pub const REVDATE: &str = env!("VCS_COMMIT_DATE");

use knockout;

use reqwest::{Client, cookie::Jar};
use std::{env, fs};
use tokio::runtime::Runtime;

use crate::knockout::Thread;

fn main() -> Result<(), eyre::Error> {
    println!("╔════════════╤═══════ KNOCKOUT SCRAPER ═══════════════════╗");
    println!("║ Version    │ {0: <43}║", VERSION);
    println!("║ Build Date │ {0: <43}║", BUILDTD);
    println!("║ Revision   │ {0: <43}║", REVISION);
    println!("║ Rev Date   │ {0: <43}║", REVDATE);
    println!("╚════════════╧════════════════════════════════════════════╝\n");

    #[cfg(target_os = "windows")]
    {
        ansi_term::enable_ansi_support()
            .map_err(|e| eyre::Report::msg(format!("Cannot enable ANSI color on Windows: {}", e)))?
    };

    // Gather passed env vars / arguments
    let mut args: Vec<String> = env::args().collect();
    // Check if there's an input.
    if args.len() < 2 {
        println!("No input!");
        println!("Usage: ");
        println!("\t List    1. knockout_search 1 5 16 8 24 (...)");
        println!("\t Range   2. knockout_search 25-800");
        return Ok(());
    };
    // If we're a range instead of a list.
    if args.len() == 2 {
        if args[1].contains(&String::from("-")) {
            let split: Vec<&str> = args[1].split('-').collect();
            let first = split[0]
                .parse::<usize>()
                .expect("First number in range could not be converted to usize!");
            let second = split[1]
                .parse::<usize>()
                .expect("Second number in range could not be converted to usize!");

            args = vec![args[0].to_owned()];
            for i in first..=second {
                args.push(i.to_string());
            }
            println!("Using Range Mode.");
            println!(
                "\tThreads: {f:width$}-{s:width$}\n",
                width = second.to_string().chars().count(),
                f = first,
                s = second,
            )
        }
    }

    let rt = Runtime::new()?;
    rt.block_on(async { real_main(args).await })?;
    Ok(())
}

async fn real_main(args: Vec<String>) -> Result<(), eyre::Error> {
    let jar = Jar::default();
    match fs::read_to_string("cookie.txt") {
        Ok(cookie) => {
            let url = knockout::API_URL.parse::<reqwest::Url>().unwrap();
            jar.add_cookie_str(&cookie, &url);
        },
        Err(_) => {},
    }
    let client: Client = {
        Client::builder()
            .gzip(true)
            .cookie_store(true)
            .cookie_provider(jar.into())
            .build()
            .expect("Trying to initalize Reqwest client for SteamAPI")
    };
    let mut skipped_ids: Vec<u64> = Vec::new();
    // Go through all of the thread IDs we're given.
    for arg in args.iter().skip(1) {
        println!(); // Extra newline.
        let thread;
        match knockout::scrape_thread(&client, arg.parse()?).await {
            Ok(res) => {
                thread = res;
            }
            Err(reason) => {
                println!("Thread:\t{} SKIPPED! \t Reason: \"{}\"", &arg, reason);
                skipped_ids.push(arg.parse()?);
                continue;
            }
        }
        // Write whole thread verbatim.
        flush_thread(thread)?;
    }
    if skipped_ids.len() > 0 {
        println!(
            "\n\
        Skipped Threads:"
        );
        for id in skipped_ids {
            print!("{} ", id);
        }
        println!("\n");
    }
    println!("Finished!");
    Ok(())
}

fn flush_thread(thread: Thread) -> Result<(), eyre::Error> {
    let json = serde_json::to_string(&thread)?;
    std::fs::create_dir_all("threads")?;
    let path = format!("threads/{}.json", thread.id);
    // Write debug file output.
    std::fs::write(path, json)?;
    Ok(())
}
